/**************************************
   Generic HID Class implementation
 -------------------------------------
 Created by David Guillen (davidgf.net)
   Have a look at microchip examples
      and the USB Complete book
 -------------------------------------
 The code is under the public domain
***************************************/

#include <pic18fregs.h>
#include "hid.h"
#include "usb.h"
#include "hid_macros.h"

unsigned char hidProtocol;
unsigned char hidIdleRate;
unsigned char HIDTransactionFlag;

#pragma udata usbram5  ep1_out_buffer ep1_in_buffer
volatile unsigned char ep1_in_buffer[HID_REPORT_SIZE];
volatile unsigned char ep1_out_buffer[HID_REPORT_SIZE];

unsigned char HIDReportDescriptor[HID_REPORT_DESC_SIZE] = {
	HID_USAGE_PAGE(HID_USAGE_GENERIC_DESKTOP),
	HID_USAGE(HID_USAGE_MOUSE),
	HID_COLLECTION(HID_COLLECTION_APP),
	HID_USAGE(HID_USAGE_POINTER),

	HID_COLLECTION(HID_COLLECTION_PHY),

	HID_USAGE_PAGE(HID_USAGE_BUTTON),
	HID_USAGE_MIN(1),HID_USAGE_MAX(8),
	HID_LOG_MIN(0),  HID_LOG_MAX(1),
	HID_REPORT_SZ(1), HID_REPORT_COUNT(8),
	HID_INPUT(2),

	HID_USAGE_PAGE(HID_USAGE_GENERIC_DESKTOP),
	HID_USAGE(HID_X_AXIS), HID_USAGE(HID_Y_AXIS),
	HID_LOG_MIN(0x81),  HID_LOG_MAX(0x7F),

	HID_REPORT_SZ(8), HID_REPORT_COUNT(2),
	HID_INPUT(6),

	HID_COLLECTION_END,
	HID_COLLECTION_END
};

// Initialize the HID subsystem (after a SET CONFIG request)
void HIDInitEndpoint(void) {
	int i;
	for (i = 0; i < HID_REPORT_SIZE; i++) {
		ep1_in_buffer[i] = 0;
		ep1_out_buffer[i] = 0;
	}

	// Turn on both in and out for this endpoint
	UEP1 = 0x1E;

	BD_PREPARE_OUT(1,HID_REPORT_SIZE,ep1_out_buffer,TDATA0)

	endps[1].in.ADDR = PTR16(ep1_in_buffer);
	endps[1].in.Stat = DTS;
	endps[1].in.Cnt = LO_B(HID_REPORT_SIZE);
}

// Process a GET_REPORT request
char HIDGetReport(void) {
	unsigned char reportID = SetupPacket.wValue0;
	
	if (SetupPacket.wValue1 == HID_IN_REPORT)        return GetInputReport(reportID);
	else if (SetupPacket.wValue1 == HID_FEAT_REPORT) return GetFeatureReport(reportID);
	return 0;
}

// Process SET_REPORT request
char HIDSetReport(void) {
	unsigned char reportID = SetupPacket.wValue0;

	if (SetupPacket.wValue1 == HID_OUT_REPORT)       return SetupOutputReport(reportID);
	else if (SetupPacket.wValue1 == HID_FEAT_REPORT) return SetupFeatureReport(reportID);
	return 0;
}

// Process HID specific requests
char ProcessHIDRequest(void) {
	unsigned char req,reqType;
	HIDTransactionFlag = 0;

	// Only process packets directed to 0 interface
	if((SetupPacket.bmRequestType & 0x1F) != RECIP_INTERFACE || (SetupPacket.wIndex0 != 0)) return 0;

	req = SetupPacket.bRequest;
	reqType = (SetupPacket.bmRequestType >> 5) & 0x3;

	switch (reqType) {
	case REQ_TYPE_STD:
		if (req == GET_DESCRIPTOR) {
			switch (SetupPacket.wValue1) {
			case HID_DESCRIPTOR:
				outPtr = (char*)&configDescriptor.HIDDescriptor;
				wCount = HID_HEADER_SIZE;
				return 1;
			case REPORT_DESCRIPTOR:
				outPtr = (char*)HIDReportDescriptor;
				wCount = HID_REPORT_DESC_SIZE;
				return 1;
			case PHYSICAL_DESCRIPTOR:
				// Physical descriptor.
				return 1;
			}
		}
	case REQ_TYPE_CLASS:
		switch (req) {
		case HID_GET_REPORT:
			return HIDGetReport();
		case HID_SET_REPORT:
			HIDTransactionFlag = 1;
			return HIDSetReport();
		case HID_GET_IDLE:
			outPtr = &hidIdleRate;
			wCount = 1;
			return 1;
		case HID_SET_IDLE:
			hidIdleRate = SetupPacket.wValue1;
			return 1;
		case HID_GET_PROTOCOL:
			outPtr = &hidProtocol;
			wCount = 1;
			return 1;
		case HID_SET_PROTOCOL:
			hidProtocol = SetupPacket.wValue0;
			return 1;
		default:
			return 0;
		};
	}
}

// An output transfer occured, check if we need to do some work
void HIDOutNotify(void) {
	if (HIDTransactionFlag) {
		unsigned char reportID = SetupPacket.wValue0;
		
		switch(SetupPacket.wValue1) {
		case HID_OUT_REPORT:
			SetOutputReport(reportID);
			return;
		case HID_FEAT_REPORT:
			SetFeatureReport(reportID);
			return;
		}
	}
}

// get report handles, setup the output buffers
char GetInputReport(unsigned char reportID) {
	outPtr = 0;
	wCount = 0;
	return 1;
}

char GetFeatureReport(unsigned char reportID) {
	outPtr = 0;
	wCount = 0;
	return 1;
}

// set report setup, prepare reception buffers
char SetupFeatureReport(unsigned char reportID) {
	inPtr = 0;
	return 1;
}

char SetupOutputReport(unsigned char reportID) {
	inPtr = 0;
	return 1;
}

// set report, check the data received
void SetFeatureReport(unsigned char reportID) {
	;
}

void SetOutputReport(unsigned char reportID) {
	;
}

