/**************************************
   Generic USB stack implementation
 -------------------------------------
 Created by David Guillen (davidgf.net)
 Based on Alexander Enzmann work, PUF
 (Gilles Ballanger & Pierre Gaufillet)
 and pic18fusb by Andrew Holbrook
 -------------------------------------
 The code is under the public domain
***************************************/

#include <pic18fregs.h>
#include "usb.h"
#include "hid.h"

volatile ENDP at 0x0400 endps[16];

unsigned char deviceState;
unsigned char remoteWakeup;
unsigned char selfPowered;
unsigned char devConfigured;
unsigned char ctrlTransferStage;

#define SETUP_STAGE    0 // Start of a control transfer (followed by 0 or more data stages)
#define DATA_OUT_STAGE 1 // Data from host to device
#define DATA_IN_STAGE  2 // Data from device to host
#define STATUS_STAGE   3 // Unused - if data I/O went ok, then back to Setup

unsigned char *outPtr;           // Data to send to the host
unsigned char *inPtr;            // Data from the host
short wCount;                    // Number of bytes of data

const unsigned char deviceDescriptor[18] = {
    0x12, 0x01, // bLength, bDescriptorType
    0x00, 0x02, // bcdUSB (low byte), bcdUSB (high byte)
    0x00, 0x00, // bDeviceClass, bDeviceSubClass
    0x00, E0SZ, // bDeviceProtocl, bMaxPacketSize
    0xD8, 0x04, // idVendor (low byte), idVendor (high byte)
    0x41, 0x45, // idProduct (low byte), idProduct (high byte)
    0x01, 0x00, // bcdDevice (low byte), bcdDevice (high byte)
    0x01, 0x02, // iManufacturer, iProduct
    0x00, 0x01  // iSerialNumber (none), bNumConfigurations
};

const unsigned char strDesc0[] = {
    sizeof(strDesc0), STRING_DESCRIPTOR,
    0x09, 0x04,
};

USB_STRING(strDesc1,CONCAT('w','w','w','.','d','a','v','i','d','g','f','.','n','e','t'))
USB_STRING(strDesc2,CONCAT('G','e','n','e','r','i','c',' ','M','O','U','S','E'))

const unsigned char * stringList[3] = {strDesc0,strDesc1,strDesc2};

code ConfigStruct configDescriptor = {
    {
    // Configuration descriptor
    CONFIG_HEADER_SIZE, CONFIGURATION_DESCRIPTOR,   // bLength, bDescriptorType (Configuration)
    LO_B(sizeof(configDescriptor)), HI_B(sizeof(configDescriptor)),
    0x01, 0x01,                                     // bNumInterfaces, bConfigurationValue
    0x00, 0xA0,                                     // iConfiguration, bmAttributes ()
    0x32,                                           // bMaxPower
    },
    {
    // HID Interface descriptor
    HID_HEADER_SIZE, INTERFACE_DESCRIPTOR,          // bLength, bDescriptorType (Interface)
    0x00, 0x00,                                     // bInterfaceNumber, bAlternateSetting
    0x01, 0x03,                                     // bNumEndpoints, bInterfaceClass (HID)
    0x00, 0x02,                                     // bInterfaceSubclass, bInterfaceProtocol, MOUSE
    0x00,                                           // iInterface
    // Hid descriptor
    0x09, 0x21,                                     // bLength, bDescriptorType (HID)
    0x01, 0x01,                                     // bcdHID (low), bcdHID (high)
    0x00, 0x01,                                     // bCountryCode, bNumDescriptors
    0x22, LO_B(HID_REPORT_DESC_SIZE),               // bDescriptorType, wDescriptorLength (low)
    HI_B(HID_REPORT_DESC_SIZE),                     // wDescriptorLength (high)
    // HID Endpoint 1 In
    0x07, 0x05,                                     // bLength, bDescriptorType (Endpoint)
    0x81, 0x03,                                     // bEndpointAddress, bmAttributes (Interrupt)
    LO_B(HID_REPORT_SIZE), HI_B(HID_REPORT_SIZE),   // wMaxPacketSize (low), wMaxPacketSize (high)
    0x0A,                                           // bInterval (10 millisecond)
    }
};

const unsigned char * buffers_table[16][2] = {	{0,0},
						{ ep1_out_buffer, ep1_in_buffer} };

const unsigned short buffers_size_table[16] = { 0, HID_REPORT_SIZE };

// EP0 buffers
#pragma udata usbram5 SetupPacket controlTransferBuffer
volatile setupPacketStruct SetupPacket;
volatile char controlTransferBuffer[E0SZ];

void CopyMemory(char * dst, char * src, unsigned short len) {
	unsigned int ci;
	for (ci = 0; ci < len; ci++)
		*dst++ = *src++;
}

// Send data to host by copying to the USB-ram buffer and setting up a transaction
unsigned char ep_send(unsigned char ep, unsigned char * buffer, unsigned short len) {
	if (endps[ep].in.Stat & UOWN) return 0;

	CopyMemory(buffers_table[ep][1], buffer, len);
	
	// TODO: Cannot send more than 255 bytes
	// no problem unless using isochronous transfers
	endps[ep].in.Cnt = len;
	if(endps[ep].in.Stat & DTS)  endps[ep].in.Stat = UOWN | DTSEN;
	else                         endps[ep].in.Stat = UOWN | DTS | DTSEN;

	return len;
}

// Reads the bytes sent by the host and sets up the buffer for a new
// incoming packet
unsigned char ep_recv(unsigned char ep, unsigned char * buffer, unsigned short len) {
	if(!(endps[ep].out.Stat & UOWN)) {
		unsigned short r_len = endps[ep].out.Cnt;
		r_len |= ((endps[ep].out.Stat&0x3)<<8);

		if (len > r_len) len = r_len; // Copy len bytes

		CopyMemory(buffer,buffers_table[ep][0],len);

		endps[ep].out.Cnt = buffers_size_table[ep];
		if(endps[ep].out.Stat & DTS) endps[ep].out.Stat = UOWN | DTSEN;
		else                         endps[ep].out.Stat = UOWN | DTS | DTSEN;

		return len;
	}
	return 0;
}


// Process GET_DESCRIPTOR
static char GetDescriptor(void) {
	if(SetupPacket.bmRequestType == 0x80) {
		switch (SetupPacket.wValue1) { // Desc. type
		case DEVICE_DESCRIPTOR:
			outPtr = deviceDescriptor;
			wCount = DEVICE_DESCRIPTOR_SIZE;
			return 1;
		case CONFIGURATION_DESCRIPTOR:
			outPtr = (char*)configDescriptor;
			wCount = sizeof(configDescriptor);
			return 1;
		case STRING_DESCRIPTOR:
			outPtr = stringList[SetupPacket.wValue0];
			wCount = *outPtr;
			return 1;
		}
	}
	return 0;
}

// Process GET_STATUS
static char GetStatus(void) {
	// Mask off the Recipient bits
	unsigned char recipient = SetupPacket.bmRequestType & 0x1F;

	controlTransferBuffer[0] = 0;
	controlTransferBuffer[1] = 0;
	outPtr = controlTransferBuffer;
	wCount = 2;

	if (recipient == 0x00) { // Device
		// Set bits for self powered device and remote wakeup.
		if (selfPowered) controlTransferBuffer[0] |= 0x01;
		if (remoteWakeup) controlTransferBuffer[0] |= 0x02;
		return 1;
	} else if (recipient == 0x01) { // Interface
		return 1;
	} else if (recipient == 0x02) { // Endpoint
		unsigned char endpointNum = SetupPacket.wIndex0 & 0x0F;
		unsigned char endpointDir = SetupPacket.wIndex0 & 0x80;

		if (endpointDir) controlTransferBuffer[0] = (endps[endpointNum].out.Stat & BSTALL);
		else             controlTransferBuffer[0] = (endps[endpointNum].in.Stat & BSTALL);
		return 1;
	}

	return 0;
}

// Process SET_FEATURE and CLEAR_FEATURE
static char SetFeature(char set) {
	unsigned char recipient = SetupPacket.bmRequestType & 0x1F;
	unsigned char feature = SetupPacket.wValue0;

	wCount = 0;
	
	if (recipient == 0x00) { // Device
		if (feature == DEVICE_REMOTE_WAKEUP) {
			remoteWakeup = set;
			return 1;
		}
	} else if (recipient == 0x02) { // Endpoint
		unsigned char endpointNum = SetupPacket.wIndex0 & 0x0F;
		unsigned char endpointDir = SetupPacket.wIndex0 & 0x80;
		if ((feature == ENDPOINT_HALT) && (endpointNum != 0)) {
			// Halt endpoint (as long as it isn't endpoint 0)
			if (set) {
				if (endpointDir)
				endps[endpointNum].out.Stat = 0x84;
				else
				endps[endpointNum].in.Stat = 0x84;
			}else{
				if (endpointDir)
				endps[endpointNum].out.Stat = 0x00;
				else
				endps[endpointNum].in.Stat = 0x88;
			}

			return 1;
		}
	}
	return 0;
}

char ProcessStandardRequest(void) {
	unsigned char request = SetupPacket.bRequest;

	if((SetupPacket.bmRequestType & 0x60) != 0x00) return 0;

	switch(request) {
	case SET_ADDRESS:
		// Wait for address change until next transaction
		deviceState = USB_ADDRESS;
		return 1;
	case GET_DESCRIPTOR:
		return GetDescriptor();
	case SET_CONFIGURATION:
		devConfigured = SetupPacket.wValue0;

		if (devConfigured == 0) {
			deviceState = USB_ADDRESS;
		}else if(devConfigured == 1){
			deviceState = USB_CONFIGURED;
			ClassInitEndpoint();
		}
		return 1;
	case GET_CONFIGURATION:
		outPtr = (char*)&devConfigured;
		wCount = 1;
		return 1;
	case GET_STATUS:
		return GetStatus();
	case CLEAR_FEATURE:
		return SetFeature(0);
	case SET_FEATURE:
		return SetFeature(1);
	case GET_INTERFACE:
		controlTransferBuffer[0] = 0;
		outPtr = controlTransferBuffer;
		wCount = 1;
		return 1;
	case SET_INTERFACE:
		return 1;
	case SET_DESCRIPTOR:
	case SYNCH_FRAME:
	default:
		return 0;
	}
}

// Data stage for a Control Transfer that sends data to the host
void InDataStage(void) {
	short bufferSize;

	// Determine how many bytes are going to the host
	if(wCount < E0SZ) bufferSize = wCount;
	else bufferSize = E0SZ;

	BD_PREPARE_IN(0,bufferSize,controlTransferBuffer,TDATA0)

	// Move data to the USB output buffer from wherever it sits now.
	CopyMemory(controlTransferBuffer,outPtr,bufferSize);
	outPtr += bufferSize;
	wCount = wCount - bufferSize;
}

// Data stage for a Control Transfer that reads data from the host
void OutDataStage(void) {
	unsigned short bufferSize = ((0x03 & endps[0].out.Stat) << 8) | endps[0].out.Cnt;

	if (inPtr != 0) {
		CopyMemory(inPtr,controlTransferBuffer,bufferSize);
		inPtr += bufferSize;
	}
	wCount = wCount + bufferSize;
}

void SetupStage(void) {
	char handled;
	endps[0].in.Stat &= ~UOWN;
	endps[0].out.Stat &= ~UOWN;
	
	// Initialize the transfer process
	ctrlTransferStage = SETUP_STAGE;
	wCount = 0;
	
	// Call the default handler and any class-specific class handler that may exist
	handled = ProcessStandardRequest();
	handled |= ProcessClassRequest();
	
	// Have a look, if we are not handling it, stall, otherwise
	// prepare the data stage
	if (!handled) { // STALL!
		endps[0].out.Cnt = E0SZ;
		endps[0].out.ADDR = PTR16(&SetupPacket);
		endps[0].out.Stat = UOWN | BSTALL;
		endps[0].in.Stat = UOWN | BSTALL;
	}else if (SetupPacket.bmRequestType & 0x80) { // Device-to-host
		if(SetupPacket.wLength < wCount)
			wCount = SetupPacket.wLength;
		InDataStage();
		ctrlTransferStage = DATA_IN_STAGE;
		// Reset the out buffer descriptor for endpoint 0
		BD_PREPARE_OUT(0,E0SZ,SetupPacket,TDATA0)
	
		endps[0].in.ADDR = PTR16(&controlTransferBuffer);
		endps[0].in.Stat = UOWN | DTS | DTSEN; 
	}else{ // Host-to-device
		ctrlTransferStage = DATA_OUT_STAGE;
	
		// Clear the input buffer descriptor
		endps[0].in.Cnt = 0;
		endps[0].in.Stat = UOWN | DTS | DTSEN;
	
		// Set the out buffer descriptor on endpoint 0 to receive data
		BD_PREPARE_OUT(0,E0SZ,controlTransferBuffer,TDATA0)
	}
	
	// Enable packet processing, was disabled by SIE
	UCONbits.PKTDIS = 0;
}

void PrepareSetupStage(void) {
	ctrlTransferStage = SETUP_STAGE;

	BD_PREPARE_OUT(0,E0SZ,SetupPacket,TDATA0)
	endps[0].in.Stat = 0x00;          // Give control to CPU
}

// The routine processes data sent to EP0
void ProcessControlTransfer(void) {	
	if (USTATbits.DIR == 0) { // Endpoint 0:out
		unsigned char PID = (endps[0].out.Stat & 0x3C) >> 2;
		if (PID == 0x0D) { // SETUP PID - a transaction is starting
			SetupStage();
		}else if (ctrlTransferStage == DATA_OUT_STAGE) {
			OutDataStage(); // Accept and process data

			ClassOutNotify(); // Notify the class about the transaction

			// Keep the EP accepting data
			if(endps[0].out.Stat & DTS) endps[0].out.Stat = UOWN | DTSEN;
			else endps[0].out.Stat = UOWN | DTS | DTSEN;
		} else {
			PrepareSetupStage(); // Error? Just restart the transaction
		}
	}else{ // Endpoint 0:in
		if ((UADDR == 0) && (deviceState == USB_ADDRESS)) {
			UADDR = SetupPacket.wValue0;
			if(UADDR == 0) deviceState = USB_DEFAULT;
		}

		if (ctrlTransferStage == DATA_IN_STAGE) {
			InDataStage();  // Start (or continue) transmitting data
			// Keep feeding data
			if(endps[0].in.Stat & DTS)  endps[0].in.Stat = UOWN | DTSEN;
			else                        endps[0].in.Stat = UOWN | DTS | DTSEN;
		}else{
			PrepareSetupStage(); // Error? Just restart the transaction
		}
	}
}


void EnableUSBModule(void) {
	if(UCONbits.USBEN == 0) {
		UCON = 0;
		UIE = 0;
		UCONbits.USBEN = 1;
		deviceState = USB_ATTACHED;
	}
	
	if ((deviceState == USB_ATTACHED) && !UCONbits.SE0) {
		UIR = 0;
		UIE = 0;
		UIEbits.URSTIE = 1;
		UIEbits.IDLEIE = 1;
		deviceState = USB_POWERED;
	}
}

// Unsuspend the device
void UnSuspend(void) {
	UCONbits.SUSPND = 0;
	UIEbits.ACTVIE = 0;
	UIRbits.ACTVIF = 0;
}

void StartOfFrame(void) {
	UIRbits.SOFIF = 0;
}

// This is called by interrupt when we stalled the EP0
void Stall(void) {
	if(UEP0bits.EPSTALL == 1) {
		// Prepare for the Setup stage of a control transfer
		PrepareSetupStage();
		UEP0bits.EPSTALL = 0;
	}
	UIRbits.STALLIF = 0;
}

// We could wait here or do something!
void Suspend(void) {
	UIEbits.ACTVIE = 1;
	UIRbits.IDLEIF = 0;
	UCONbits.SUSPND = 1;
}

void BusReset() {
	UEIR  = 0x00;
	UIR   = 0x00;
	UEIE  = 0x9f;
	UIE   = 0x7b;
	UADDR = 0x00;
	
	UEP0 = 0x16;
	
	// Flush FIFO
	while (UIRbits.TRNIF == 1) UIRbits.TRNIF = 0;

	// Prepare for setup stage	
	UCONbits.PKTDIS = 0;
	PrepareSetupStage();
	
	remoteWakeup = 0;
	selfPowered = 0;
	devConfigured = 0;
	deviceState = USB_DEFAULT;
}

void ProcessTransfer(unsigned char endp) {
	// Transfer to other EPs
	// HID uses interrupts, nothing to do!
}

// Main entry point for USB tasks. Checks interrupts, then checks for transactions.
void ProcessUSBTransactions(void) {
	if(deviceState == USB_DETACHED) return;
	
	if(UIRbits.ACTVIF && UIEbits.ACTVIE) UnSuspend();

	if(UCONbits.SUSPND == 1) return;
	
	if (UIRbits.URSTIF && UIEbits.URSTIE)   BusReset();
	if (UIRbits.IDLEIF && UIEbits.IDLEIE)   Suspend();
	if (UIRbits.SOFIF && UIEbits.SOFIE)     StartOfFrame();
	if (UIRbits.STALLIF && UIEbits.STALLIE) Stall();
	if (UIRbits.UERRIF && UIEbits.UERRIE)   UIRbits.UERRIF = 0;

	if (deviceState < USB_DEFAULT) return;

	// A transaction has finished.  Try default processing on endpoint 0.
	if(UIRbits.TRNIF && UIEbits.TRNIE) {
		if (USTATbits.ENDP == 0)
			ProcessControlTransfer();
		else
			ProcessTransfer(USTATbits.ENDP);
		// Turn off interrupt and update the FIFO
		UIRbits.TRNIF = 0;
	}
}

// Full-speed with internal pullups
void InitializeUSB() {
	UCFG = 0x14;
	deviceState = USB_DETACHED;
	remoteWakeup = 0;
	devConfigured = 0;
	inPtr = 0;
	wCount = 0;
}
